package org.ibex.xt;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.ibex.util.*;
import org.ibex.util.Collections;
import org.ibex.js.*;

import org.prevayler.*;
import org.prevayler.implementation.snapshot.*;

public class Servlet extends HttpServlet {

    private String path;
    private Prevayler prevayler;
    private JS prevalent;
    private ServletContext cx = null;

    public void destroy() { try {
        synchronized(this.getClass()) {
            Prevayler privatePrevayler = prevayler;
            if (prevayler == null) return;
            prevayler = null;
            Prevalence.destroy(cx, prevayler);
        }
    } catch (Exception e) { e.printStackTrace(); } }

    public void init(ServletConfig sc) throws ServletException {
        cx = sc.getServletContext();
        prevayler = Prevalence.getPrevayler(cx);
        prevalent = (JS)prevayler.prevalentSystem();
    }

    public void doPost(HttpServletRequest request, HttpServletResponse response) throws IOException { doGet(request, response); }
    public void doGet(HttpServletRequest rq, HttpServletResponse rs) throws IOException {
        String path = cx.getRealPath(rq.getServletPath());
        Servlet.Scope scope = new Servlet.Scope(cx, rq, rs, prevayler);
        rs.getWriter().write("beginning output...");
        try { Template.wrap(Template.parse(path, scope), scope).out(rs.getWriter()); }
        catch (Exception e) { e.printStackTrace(); System.out.println("e = "+e); }
        rs.getWriter().write("...output ends.");
    }

    public static class Scope extends Template.Scope {
        private final ServletContext cx;
        private final HttpServletRequest request;
        private final HttpServletResponse response;
        private final Prevayler prevayler;

        public Scope(ServletContext cx, HttpServletRequest rq, HttpServletResponse rs, Prevayler p) {
            super(null); this.cx = cx; request = rq; response = rs; prevayler = p;
        }

        public String getLocalPath() { return cx.getRealPath("/") + "/WEB-INF/"; }
        public void transaction(JS t) {
            try { prevayler.execute(new Prevalence.JSTransaction(t)); }
            catch (Exception e) { e.printStackTrace(); throw new RuntimeException(e); }
        }

        private JS session = new JS() {
            public Object get(Object key) throws JSExn {
                //#switch(JS.toString(key))
                case "created":       return new JSDate(request.getSession(true).getCreationTime());
                case "accessed":      return new JSDate(request.getSession(true).getLastAccessedTime());
                case "invalidate":    return METHOD;
                //#end
                return super.get(key);
            }
            public void put(Object key, Object val) throws JSExn {
                //#switch(JS.toString(key))
                case "created":       throw new JSExn("can not set session.created");
                case "accessed":      throw new JSExn("can not set session.accessed");
                case "invalidate":    throw new JSExn("can not set session.invalidate");
                //#end
                super.put(key, val);
            }
            public Object callMethod(Object method, final Object a, final Object b, Object c, Object[] rest, int nargs)
                throws JSExn {
                //#switch(JS.toString(method))
                case "invalidate":    request.getSession(true).invalidate(); return null;
                //#end
                return super.callMethod(method, a, b, c, rest, nargs);
            }
        };
        private JS params = new JS() {
            private List keys = null;
            public Object get(Object key) { return request.getParameter(JS.toString(key)); }
            public Collection keys() {
                return keys == null ? keys = Collections.list(request.getParameterNames()) : keys; }
        };
        private JS cookies = new JS() {
                /*
                public Object get(Object key) { return request.getCookie(JS.toString(key)); }
                public Enumeration keys() { return request.getCookieNames(); }
                */
            };
        private JS sessionAttributes = new JS() {
            private List keys = null;
            public Object get(Object key) {
                return request.getSession(true).getAttribute(JS.toString(key)); }
            public void put(Object key, Object val) {
                if (val == null) request.getSession(true).removeAttribute(JS.toString(key));
                else request.setAttribute(JS.toString(key), val); }
            public Collection keys() {
                return keys == null ? keys = Collections.list(request.getSession(true).getAttributeNames()) : keys; }
        };
        private JS requestHeader = new JS() {
            private List keys = null;
            public Object get(Object key) { return request.getHeader(JS.toString(key)); }
            public Collection keys() {
                return keys == null ? keys = Collections.list(request.getHeaderNames()) : keys; }
        };
        private JS responseHeader = new JS() {
            public void put(Object key, Object val) {
                response.setHeader(JS.toString(key), JS.toString(val)); }
        };


        /** lets us put multi-level get/put/call keys all in the same method */
        private class Sub extends JS {
            Object key;
            Sub(Object key) { this.key = key; }
            public void put(Object key, Object val) throws JSExn {
                Scope.this.put(JS.toString(this.key) + "." + JS.toString(key), val); }
            public Object get(Object key) throws JSExn {
                return Scope.this.get(JS.toString(this.key) + "." + JS.toString(key)); }
            public Object call(Object a0, Object a1, Object a2, Object[] rest, int nargs) throws JSExn {
                return Scope.this.callMethod(this.key, a0, a1, a2, rest, nargs);
            }
            public Object callMethod(Object method, Object a0, Object a1, Object a2, Object[] rest, int nargs) throws JSExn {
                return Scope.this.callMethod(JS.toString(this.key) + "."
                                                         + JS.toString(method), a0, a1, a2, rest, nargs);
            }
        }
        private Sub getSub(String key) { return new Sub(key); }

        public Object callMethod(Object method, final Object a, final Object b, Object c, Object[] rest, int nargs) throws JSExn {
            //#switch(method)
            case "session.invalidate":    request.getSession(true).invalidate(); return null;
            case "context.list":
                String path = JS.toString(a);
                if (path.indexOf("..") != -1) throw new JSExn("cannot use .. in paths");
                File f = new File(cx.getRealPath("/") + File.separatorChar + path);
                if (!f.isDirectory()) return null;
                String[] contents = f.list();
                JSArray ret = new JSArray(contents.length);
                for(int i=0; i<contents.length; i++) ret.add(contents[i]);
                return ret;
            //#end
            return null;
        }
        public Object get(Object key) throws JSExn {
            //#switch(key)
            case "body":
            case "arg":                   return null;
            case "prevalent":             return prevayler.prevalentSystem();
            case "request":               return getSub("request");
            case "request.user":          return request.getRemoteUser();
            case "request.header":        return requestHeader;
            case "request.method":        return request.getMethod();
            case "request.remote":        return getSub("request.remote");
            case "request.remote.ip":     return request.getRemoteAddr();
            case "request.remote.host":   return request.getRemoteHost();
            case "request.ssl":           return new Boolean(request.isSecure());
            case "request.path":          return request.getPathInfo();
            case "response":              return getSub("response");
            case "response.header":       return responseHeader;
            case "session":               return session;
            case "page":                  return getSub("page");
            //case "page.lastmodified":     return new JSDate(new File(path).lastModified()); FIXME
            case "context":               return getSub("context");
            case "context.list":          return METHOD;
            case "params":                return params;
            case "cookie":                return cookies;
            case "xt.date":               return new JSDate(); // TODO: discuss
            //#end
            return null;
        }
        public void put(Object key, Object val) throws JSExn {
            try {
            //#switch(JS.toString(key))
            case "transaction":           transaction((JS)val);
            case "response.code":         response.setStatus(JS.toInt(val));
            case "response.redirect":     response.sendRedirect(JS.toString(val));
            case "response.contentType":  response.setContentType(JS.toString(val));
            //#end
            } catch (IOException e) {
                throw new JSExn(e);
            }
        }
    }

}
